@@ -29,6 +29,8 @@
 
 #include <gdk/gdk.h>
 #include <gdk/gdkx.h>
+#include <gdk-pixbuf/gdk-pixbuf.h>
+#include <cairo.h>
 
 #include "terminal-accels.h"
 #include "terminal-app.h"
@@ -69,6 +71,8 @@ struct _TerminalScreenPrivate
 	gboolean user_title; /* title was manually set */
 	GSList *match_tags;
 	guint launch_child_source_id;
+	gulong bg_image_callback_id;
+	GdkPixbuf *bg_image;
 };
 
 enum
@@ -226,7 +230,8 @@ cwd_of_pid (int pid)
 				if (chdir (cwd_file) == 0)
 				{
 					working_dir = g_get_current_dir ();
-					(void) chdir (cwd);
+					if (chdir (cwd) < 0)
+						g_warning ("Could not change working directory.");
 				}
 				g_free (cwd);
 			}
@@ -386,6 +391,9 @@ terminal_screen_init (TerminalScreen *screen)
 	g_signal_connect (terminal_app_get (), "notify::system-font",
 	                  G_CALLBACK (terminal_screen_system_font_notify_cb), screen);
 
+	priv->bg_image_callback_id = 0;
+	priv->bg_image = NULL;
+
 #ifdef MATE_ENABLE_DEBUG
 	_TERMINAL_DEBUG_IF (TERMINAL_DEBUG_GEOMETRY)
 	{
@@ -662,9 +670,56 @@ terminal_screen_finalize (GObject *object)
 	g_slist_foreach (priv->match_tags, (GFunc) free_tag_data, NULL);
 	g_slist_free (priv->match_tags);
 
+	if (priv->bg_image)
+		g_object_unref (priv->bg_image);
+
 	G_OBJECT_CLASS (terminal_screen_parent_class)->finalize (object);
 }
 
+static gboolean
+terminal_screen_image_draw_cb (GtkWidget *widget, cairo_t *cr, void *userdata)
+{
+	TerminalScreen *screen = TERMINAL_SCREEN (widget);
+	TerminalScreenPrivate *priv = screen->priv;
+	GdkPixbuf *bg_image = priv->bg_image;
+	GdkRectangle target_rect;
+	GtkAllocation alloc;
+	cairo_surface_t *child_surface;
+	cairo_t *child_cr;
+
+	if (!bg_image)
+		return FALSE;
+
+	gtk_widget_get_allocation (widget, &alloc);
+
+	target_rect.x = 0;
+	target_rect.y = 0;
+	target_rect.width = alloc.width;
+	target_rect.height = alloc.height;
+
+	child_surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, alloc.width, alloc.height);
+	child_cr = cairo_create (child_surface);
+
+	g_signal_handler_block (screen, priv->bg_image_callback_id);
+	gtk_widget_draw (widget, child_cr);
+	g_signal_handler_unblock (screen, priv->bg_image_callback_id);
+
+	gdk_cairo_set_source_pixbuf (cr, bg_image, 0, 0);
+	cairo_pattern_set_extend (cairo_get_source (cr), CAIRO_EXTEND_REPEAT);
+
+	gdk_cairo_rectangle (cr, &target_rect);
+	cairo_fill (cr);
+
+	cairo_set_source_surface (cr, child_surface, 0, 0);
+	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
+	cairo_paint (cr);
+
+	cairo_destroy (child_cr);
+	cairo_surface_destroy (child_surface);
+
+	return TRUE;
+}
+
 TerminalScreen *
 terminal_screen_new (TerminalProfile *profile,
                      char           **override_command,
@@ -930,6 +985,7 @@ terminal_screen_profile_notify_cb (TerminalProfile *profile,
 	        prop_name == I_(TERMINAL_PROFILE_BACKGROUND_COLOR) ||
 	        prop_name == I_(TERMINAL_PROFILE_BACKGROUND_TYPE) ||
 	        prop_name == I_(TERMINAL_PROFILE_BACKGROUND_DARKNESS) ||
+	        prop_name == I_(TERMINAL_PROFILE_BACKGROUND_IMAGE) ||
 	        prop_name == I_(TERMINAL_PROFILE_BOLD_COLOR_SAME_AS_FG) ||
 	        prop_name == I_(TERMINAL_PROFILE_BOLD_COLOR) ||
 	        prop_name == I_(TERMINAL_PROFILE_PALETTE))
@@ -937,12 +993,9 @@ terminal_screen_profile_notify_cb (TerminalProfile *profile,
 
 	if (!prop_name || prop_name == I_(TERMINAL_PROFILE_SILENT_BELL))
 		vte_terminal_set_audible_bell (vte_terminal, !terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_SILENT_BELL));
-
-#if VTE_CHECK_VERSION (0, 40, 0)
 	if (!prop_name || prop_name == I_(TERMINAL_PROFILE_WORD_CHARS))
 		vte_terminal_set_word_char_exceptions (vte_terminal,
 		                                       terminal_profile_get_property_string (profile, TERMINAL_PROFILE_WORD_CHARS));
-#endif
 	if (!prop_name || prop_name == I_(TERMINAL_PROFILE_SCROLL_ON_KEYSTROKE))
 		vte_terminal_set_scroll_on_keystroke (vte_terminal,
 		                                      terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_SCROLL_ON_KEYSTROKE));
@@ -1014,16 +1067,26 @@ update_color_scheme (TerminalScreen *screen)
 	TerminalProfile *profile = priv->profile;
 	GdkRGBA colors[TERMINAL_PALETTE_SIZE];
 	const GdkRGBA *fg_rgba, *bg_rgba, *bold_rgba;
+	TerminalBackgroundType bg_type;
+	const gchar *bg_image_file;
 	double bg_alpha = 1.0;
 	GdkRGBA fg, bg;
+	GdkRGBA *c;
 	guint n_colors;
 	GtkStyleContext *context;
+	GError *error = NULL;
 
 	context = gtk_widget_get_style_context (GTK_WIDGET (screen));
 	gtk_style_context_save (context);
 	gtk_style_context_set_state (context, GTK_STATE_FLAG_NORMAL);
 	gtk_style_context_get_color (context, GTK_STATE_FLAG_NORMAL, &fg);
-	gtk_style_context_get_background_color (context, GTK_STATE_FLAG_NORMAL, &bg);
+
+	gtk_style_context_get (context, GTK_STATE_FLAG_NORMAL,
+			       GTK_STYLE_PROPERTY_BACKGROUND_COLOR,
+			       &c, NULL);
+	bg = *c;
+	gdk_rgba_free (c);
+
 	gtk_style_context_restore (context);
 
 	bold_rgba = NULL;
@@ -1045,10 +1108,37 @@ update_color_scheme (TerminalScreen *screen)
 	n_colors = G_N_ELEMENTS (colors);
 	terminal_profile_get_palette (priv->profile, colors, &n_colors);
 
-	if (terminal_profile_get_property_enum (profile, TERMINAL_PROFILE_BACKGROUND_TYPE) == TERMINAL_BACKGROUND_TRANSPARENT)
+	bg_type = terminal_profile_get_property_enum (profile, TERMINAL_PROFILE_BACKGROUND_TYPE);
+	bg_image_file = terminal_profile_get_property_string (profile, TERMINAL_PROFILE_BACKGROUND_IMAGE_FILE);
+
+	if (bg_type == TERMINAL_BACKGROUND_TRANSPARENT)
 		bg_alpha = terminal_profile_get_property_double (profile, TERMINAL_PROFILE_BACKGROUND_DARKNESS);
+	else if (bg_type == TERMINAL_BACKGROUND_IMAGE)
+	  bg_alpha = 0.0;
 	bg.alpha = bg_alpha;
 
+	if (bg_type == TERMINAL_BACKGROUND_IMAGE)
+	{
+		if (!priv->bg_image_callback_id)
+			priv->bg_image_callback_id = g_signal_connect (screen, "draw", G_CALLBACK (terminal_screen_image_draw_cb), NULL);
+
+		g_clear_object (&priv->bg_image);
+		priv->bg_image = gdk_pixbuf_new_from_file (bg_image_file, &error);
+
+		if (error) {
+			g_printerr ("Failed to load background image: %s\n", error->message);
+			g_clear_error (&error);
+		}
+
+		gtk_widget_queue_draw (GTK_WIDGET (screen));
+	} else {
+		if (priv->bg_image_callback_id)
+		{
+			g_signal_handler_disconnect (screen, priv->bg_image_callback_id);
+			priv->bg_image_callback_id = 0;
+		}
+	}
+
 	vte_terminal_set_colors (VTE_TERMINAL (screen),
 	                         &fg, &bg,
 	                         colors, n_colors);
@@ -1338,7 +1428,6 @@ get_child_environment (TerminalScreen *screen,
 	g_hash_table_remove (env_table, "LINES");
 	g_hash_table_remove (env_table, "MATE_DESKTOP_ICON");
 
-	g_hash_table_replace (env_table, g_strdup ("COLORTERM"), g_strdup (EXECUTABLE_NAME));
 	g_hash_table_replace (env_table, g_strdup ("TERM"), g_strdup ("xterm")); /* FIXME configurable later? */
 
 	/* FIXME: moving the tab between windows, or the window between displays will make the next two invalid... */
@@ -1405,6 +1494,49 @@ info_bar_response_cb (GtkWidget *info_bar,
 	}
 }
 
+static void handle_error_child (TerminalScreen *screen,
+				GError         *err)
+{
+	GtkWidget *info_bar;
+
+	info_bar = terminal_info_bar_new (GTK_MESSAGE_ERROR,
+	                                  _("_Profile Preferences"), RESPONSE_EDIT_PROFILE,
+	                                  _("_Relaunch"), RESPONSE_RELAUNCH,
+	                                  NULL);
+	terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
+	                               _("There was an error creating the child process for this terminal"));
+	terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
+	                               "%s", err->message);
+	g_signal_connect (info_bar, "response",
+	                  G_CALLBACK (info_bar_response_cb), screen);
+
+	gtk_box_pack_start (GTK_BOX (terminal_screen_container_get_from_screen (screen)),
+	                    info_bar, FALSE, FALSE, 0);
+	gtk_info_bar_set_default_response (GTK_INFO_BAR (info_bar), GTK_RESPONSE_CANCEL);
+	gtk_widget_show (info_bar);
+
+	g_error_free (err);
+}
+
+static void term_spawn_callback (GtkWidget *terminal,
+				 GPid       pid,
+				 GError    *error,
+				 gpointer   user_data)
+{
+	TerminalScreen *screen = TERMINAL_SCREEN (terminal);
+
+	if (error)
+	{
+		handle_error_child (screen, error);
+		g_error_free (error);
+	}
+	else
+	{
+		TerminalScreenPrivate *priv = screen->priv;
+		priv->child_pid = pid;
+	}
+}
+
 static gboolean
 terminal_screen_launch_child_cb (TerminalScreen *screen)
 {
@@ -1417,7 +1549,6 @@ terminal_screen_launch_child_cb (TerminalScreen *screen)
 	const char *working_dir;
 	VtePtyFlags pty_flags = VTE_PTY_DEFAULT;
 	GSpawnFlags spawn_flags = 0;
-	GPid pid;
 
 	priv->launch_child_source_id = 0;
 
@@ -1439,36 +1570,9 @@ terminal_screen_launch_child_cb (TerminalScreen *screen)
 	if (!terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_UPDATE_RECORDS))
 		pty_flags |= VTE_PTY_NO_UTMP | VTE_PTY_NO_WTMP;
 
-	if (!get_child_command (screen, shell, &spawn_flags, &argv, &err) ||
-	        !vte_terminal_spawn_sync (
-	                terminal,
-	                pty_flags,
-	                working_dir,
-	                argv,
-	                env,
-	                spawn_flags,
-	                NULL, NULL,
-	                &pid,
-	                NULL,
-	                &err))
+	if (!get_child_command (screen, shell, &spawn_flags, &argv, &err))
 	{
-		GtkWidget *info_bar;
-
-		info_bar = terminal_info_bar_new (GTK_MESSAGE_ERROR,
-		                                  _("_Profile Preferences"), RESPONSE_EDIT_PROFILE,
-		                                  _("_Relaunch"), RESPONSE_RELAUNCH,
-		                                  NULL);
-		terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
-		                               _("There was an error creating the child process for this terminal"));
-		terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
-		                               "%s", err->message);
-		g_signal_connect (info_bar, "response",
-		                  G_CALLBACK (info_bar_response_cb), screen);
-
-		gtk_box_pack_start (GTK_BOX (terminal_screen_container_get_from_screen (screen)),
-		                    info_bar, FALSE, FALSE, 0);
-		gtk_info_bar_set_default_response (GTK_INFO_BAR (info_bar), GTK_RESPONSE_CANCEL);
-		gtk_widget_show (info_bar);
+		handle_error_child (screen, err);
 
 		g_error_free (err);
 		g_strfreev (env);
@@ -1477,7 +1581,19 @@ terminal_screen_launch_child_cb (TerminalScreen *screen)
 		return FALSE;
 	}
 
-	priv->child_pid = pid;
+        vte_terminal_spawn_async (terminal,
+				  pty_flags,
+				  working_dir,
+				  argv,
+				  env,
+				  spawn_flags,
+				  NULL,
+				  NULL,
+				  NULL,
+				  -1,
+				  NULL,
+				  (VteTerminalSpawnAsyncCallback) term_spawn_callback,
+				  NULL);
 
 	g_free (shell);
 	g_strfreev (argv);
